"""
Abstract websocket functionality.

"""

from __future__ import print_function
from __future__ import unicode_literals

from base64 import b64encode
from hashlib import sha1
import logging
import os

from six.moves.urllib.parse import urlparse

from . import constants
from . import errors
from . import events
from .frame import Frame
from .opcode import Opcode
from .response import Response
from .stream import WebsocketStream
from .session import WebsocketSession
from .status import Status


log = logging.getLogger('ws')


class WebSocket(object):
    """IO independent websocket functionality."""

    def __init__(self, url, protocols=None):
        self.url = url
        self.protocols = protocols or []
        self._session = None

        self.stream = WebsocketStream()
        _url = urlparse(url)
        self.scheme = _url.scheme
        print(_url)
        host, _, port = _url.netloc.partition(':')
        if not port:
            port = '443' if self.scheme == 'wss' else '80'
        if not port.isdigit():
            raise ValueError('illegal port value')
        port = int(port)
        self.host = host
        self.port = port
        self._host_port = "{}:{}".format(host, port)
        self.resource = _url.path or '/'
        if _url.query:
            self.resource = "{}?{}".format(host, _url.query)
        self.key = b64encode(os.urandom(16))

        self._sent_request = False
        self._running = False
        self._closing = False
        self._closed = False

    def __repr__(self):
        return "WebSocket('{}')".format(self.url)

    @property
    def session(self):
        return self._session

    def connect(self, poll=5, session_class=WebsocketSession):
        if self._running:
            raise errors.WebSocketInUse(
                "Can't connect while WebSocket is running"
            )
        self._session = WebsocketSession(self)
        self._running = True
        return self._session.events(poll=poll)

    __iter__ = connect

    def close(self, code=Status.NORMAL, reason=b'goodbye'):
        self._send_close(code, reason)
        self._closing = True

    @property
    def is_closing(self):
        return self._closing

    @property
    def is_closed(self):
        return self._closed

    def feed(self, data):
        """Feed with data from the socket."""
        session = self.session
        for message in self.stream.feed(data):
            log.debug('%r', message)
            if isinstance(message, Response):
                try:
                    yield self.on_response(message)
                except errors.HandshakeError as error:
                    yield events.Rejected(str(error))
                    break
                except Exception as error:
                    log.exception('on_response failed')
                    yield events.Rejected(str(error))
                    break
            else:
                if message.is_close:
                    if self._closing:
                        yield events.Closed(message.code, message.reason)
                        self._closed = True
                    else:
                        self.close(message.code, message.reason)
                        self._closing = True
                elif message.is_ping:
                    session.write(Opcode.PONG, message.payload)
                elif message.is_pong:
                    yield events.Pong(message)
                elif message.is_binary:
                    yield events.Binary(message)
                elif message.is_text:
                    yield events.Text(message)
                else:
                    yield events.UnknownMessage(message)

    def get_request(self):
        """Get the request (in bytes)"""
        request = [
            "GET {} HTTP/1.1".format(self.resource)
        ]
        protocols = ", ".join(self.protocols)
        version = '{}'.format(constants.WS_VERSION)
        headers = [
            ('Host', self._host_port),
            ('Upgrade', 'websocket'),
            ('Connection', 'Upgrade'),
            ('Sec-WebSocket-Protocol', protocols),
            ('Sec-WebSocket-Key', self.key),
            ('Sec-WebSocket-Version', version)
        ]
        for header, value in headers:
            request.append('{}: {}'.format(header, value).encode())
        request.append('\r\n')
        request_bytes = b'\r\n'.join(line.encode() for line in request)
        return request_bytes

    def on_response(self, response):
        """Called when the HTTP response has been received."""

        if response.status_code != 101:
            raise errors.HandshakeError(
                'Websocket upgrade failed (code={})',
                response.status_code
            )

        upgrade_header = response.get(b'upgrade', b'?').lower()
        if upgrade_header != b'websocket':
            raise errors.HandshakeError(
                "Can't upgrade to {}",
                upgrade_header.decode(errors='replace')
            )

        accept_header = response.get(b'sec-websocket-accept', None)
        if accept_header is None:
            raise errors.HandshakeError(
                "No Sec-WebSocket-Accept header"
            )

        challenge = b64encode(
            sha1(self.key + constants.WS_KEY).digest()
        )

        if accept_header.lower() != challenge.lower():
            raise errors.HandshakeError(
                "Web-WebSocket-Accept challenge failed"
            )

        protocol = response.get(b'sec-websocket-protocol')
        extensions = set(response.get_list(b'sec-websocket-extensions') or [])

        return events.Accepted(protocol, extensions)

    def send_ping(self, data):
        """Send a ping request."""
        if len(data) <= 125:
            raise ValueError('ping data should be <= 125')
        self.session.send(Opcode.PING, data)

    def send_pong(self, data):
        """Send a ping request."""
        if len(data) <= 125:
            raise ValueError('ping data should be <= 125')
        self.session.send(Opcode.PONG, data)

    def send_binary(self, data):
        """Send a binary frame."""
        self.session.send(Opcode.BINARY, data)

    def send_text(self, text):
        """Send a text frame."""
        self.session.send(Opcode.TEXT, text.encode(errors='replace'))

    def _send_close(self, code, reason):
        """Send a close frame."""
        frame_bytes = Frame.build_close_payload(code, reason)
        self.session.send(Opcode.CLOSE, frame_bytes)

    def send_request(self):
        """Send the HTTP request."""
        request = self.get_request()
        self.session.write(request)
        self._sent_request = True



from __future__ import unicode_literals

import time


class Event(object):
    """Base class for a websocket 'event'."""
    def __init__(self):
        self.received_time = time.time()

    def __repr__(self):
        return "{}()".format(self.__class__.__name__)


class Poll(Event):
    """A generated poll event."""


class Connecting(Event):
    """Connection process has started."""

    def __init__(self, url):
        self.url = url
        super(Connecting, self).__init__()

    def __repr__(self):
        return "{}('{}')".format(self.__class__.__name__, self.url)


class Connected(Connecting):
    """Connected to server (but not yet received response)."""


class ConnectFail(Event):
    """Connection failed (connectivity related)."""
    def __init__(self, reason):
        self.reason = reason
        super(ConnectFail, self).__init__()

    def __repr__(self):
        return '{}({!r})'.format(
            self.__class__.__name__,
            self.reason,
        )


class Accepted(Event):
    """Server accepted WS connection."""
    def __init__(self, protocol, extensions):
        self.protocol = protocol
        self.extensions = extensions
        super(Accepted, self).__init__()

    def __repr__(self):
        return '{}(protocol={!r}, extensions={!r})'.format(
            self.__class__.__name__,
            self.protocol,
            self.extensions
        )


def Rejected(Event):
    """Server rejected WS connection."""
    def __init__(self, reason):
        self.reason = reason
        super(Rejected, self).__init__()

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, self.reason)


class Disconnected(Event):
    """Server disconnected."""


class Closed(Event):
    """Websocket connection is closed."""
    def __init__(self, code, reason):
        self.code = code
        self.reason = reason
        super(Closed, self).__init__()

    def __repr__(self):
        return '{}({!r}, {!r})'.format(
            self.__class__.__name__,
            self.code,
            self.reason,
        )

class UnknownMessage(Event):
    """
    An application message was received, with an unknown
    opcode.
    """
    def __init__(self, message):
        self.message = message
        super(UnknownMessage, self).__init__()


class Binary(Event):
    """An application message was received."""
    def __init__(self, message):
        self.data = message.data
        super(Binary, self).__init__()

    def __repr__(self):
        return "{}({!r})".format(self.__class__.__name__, self.data)

class Pong(Event):
    """An application message was received."""
    def __init__(self, message):
        self.data = message.data
        super(Pong, self).__init__()

    def __repr__(self):
        return "{}({!r})".format(self.__class__.__name__, self.data)

class Text(Event):
    """An application text message was received."""
    def __init__(self, message):
        self.text = message.text
        super(Text, self).__init__()

    def __repr__(self):
        return "{}({!r})".format(self.__class__.__name__, self.text)






class Status:
    NORMAL = 1000
    GOING_AWAY = 1001
    DATA_ERROR = 1002

    RESERVED = 1004
    RESERVED2 = 1005

    BAD_DATA = 1007
    POLICY_VIOLATION = 1008
    MESSAGE_TOO_LARGE = 1009
    EXTENSION_FAILED = 1010

    UNEXPECTED_CONDITION = 1011



from __future__ import print_function
from __future__ import unicode_literals

from collections import deque
from .frame import Frame
import select
import socket
import threading
import time

import six


from . import events


class WebsocketSession(object):

    def __init__(self, websocket, reconnect=True):
        self.websocket = websocket
        self.reconnect = reconnect
        self._address = (websocket.host, websocket.port)
        self._lock = threading.Lock()

        self._sock = None
        self.url = websocket.url
        self._write_buffer = deque()
        self._sent_close = False

    def __repr__(self):
        return "<ws-session '{}'>".format(self.url)

    def write(self, data, flush=True):
        """Send raw data."""
        # TODO: split large packets?
        self._write_buffer.append(data)
        if flush:
            self.flush()

    def send(self, opcode, data):
        """Send a WS Frame."""
        frame = Frame(opcode, payload=data)
        self.write(frame.to_bytes())

    def _select(self, sock, poll):
        return select.select(
            [sock],
            [sock] if self._write_buffer else [],
            [sock],
            poll
        )

    def flush(self):
        with self._lock:
            sock = self._sock
            while self._write_buffer:
                data = self._write_buffer.popleft()
                sent = sock.send(data)
                if sent != len(data):
                    self._write_buffer.appendleft(data[sent:])

    def _make_socket(self):
        sock = socket.socket(
            socket.AF_INET, socket.SOCK_STREAM
        )
        return sock

    def events(self, poll=15):
        # TODO: implement exponential back off
        websocket = self.websocket

        while 1:
            yield events.Connecting(websocket.url)
            try:
                sock = self._sock = self._make_socket()
                sock.connect(self._address)
            except socket.error as error:
                yield events.ConnectFail(six.text_type(error))
                time.sleep(5)
                continue
            else:
                break

        yield events.Connected(websocket.url)
        websocket.send_request()

        poll_start = time.time()
        while not websocket.is_closed:
            try:
                reads, writes, errors = self._select(sock, poll)
            except KeyboardInterrupt:
                websocket.close(1000, 'goodbye')
                continue
            if writes:
                self.flush()
            if reads:
                try:
                    data = sock.recv(4096)
                except socket.error:
                    break
                if not data:
                    break
                for event in websocket.feed(data):
                    yield event
            if errors:
                break
            current_time = time.time()
            if current_time - poll_start > poll:
                poll_start = current_time
                yield events.Poll()

        yield events.Disconnected()
        try:
            sock.shutdown(socket.SHUT_RDWR)
            sock.close()
        except socket.error:
            pass
        else:
            sock = None


if __name__ == "__main__":

    # Test with wstest -m echoserver -w ws://127.0.0.1:9001 -d
    # Get wstest app from http://autobahn.ws/testsuite/

    from .websocket import WebSocket

    ws = WebSocket('ws://127.0.0.1:9001')
    for event in ws.connect(poll=5):
        print(event)
        if isinstance(event, events.Poll):
            ws.send_text('Hello, World')
            ws.send_binary(b'hello world in binary')

